> module Queue where

> import Data.Sequence
> import Control.Arrow
> import Control.Monad

> type Queue a = Seq a

> emptyQ           :: Queue a
> emptyQ           =  empty

> enQueue          :: a -> Queue a -> Queue a
> enQueue          =  (<|)

> deQueue          :: Queue a -> Maybe (Queue a, a)
> deQueue q        =  case viewr q of
>                       EmptyR   -> Nothing
>                       (q :> a) -> Just (q, a)

> rollQueue        :: a -> Queue a -> (Queue a, a)
> rollQueue a q    =  case viewr q of
>                       EmptyR   -> (emptyQ, a)
>                       (q :> b) -> (a <| q, b) 

> deQueueIf        :: (a -> Bool) -> Queue a -> Maybe (Queue a, a)
> deQueueIf p q    =  do (q', a) <- deQueue q
>                        guard (p a)
>                        return (q', a) 


-- The first element out is the head of the result list

> deQueueWhile     :: (a -> Bool) -> Queue a -> (Queue a, [a])
> deQueueWhile p q =  case deQueueIf p q of
>                       Nothing       -> (q, [])
>                       Just (q', a)  -> second (a:) (deQueueWhile p q')

-- deQueueWhileR reverses the order, i.e. the last element out is the head of the list

> deQueueWhileR     :: (a -> Bool) -> Queue a -> (Queue a, [a])
> deQueueWhileR p q =  deQueueWhileRAux p q []
>   where
>     deQueueWhileRAux         :: (a -> Bool) -> Queue a -> [a] -> (Queue a, [a])
>     deQueueWhileRAux p q as  =  case deQueueIf p q of
>                                   Nothing      -> (q, as)
>                                   Just (q', a) -> deQueueWhileRAux p q' (a : as)

-------------------------------------------------------------------------
